## Tiger runtime library ------------------------------------------------------
# Based on code taken from:
# http://www.math.tau.ac.il/~nurr/courses/compile99/runtime.s
#
# entry point in SPIM: 0x004000dc
# main then jumps to t_main generated by the compiler.
#

## Routine: alloc ------------------------------------------
# Makes sure the allocation is aligned on 4.
# This routine explicitly preserves $t0 to $t7 and $a0-$a3.
# $a0 holds the size of memory wanted.
	.text
alloc:
	## $t9 holds wanted memory size
	move	$t9, $a0
	## Check that $a0 is a multiple of 4
	li	$t8, 4
	remu	$t8, $a0, $t8
	beq	$zero, $t8, L_alloc_sbrk
	## Force $a0 to be a multiple of 4
	sub	$a0, $a0, $t8
	addi	$a0, $a0, 4
L_alloc_sbrk:
	## Call sbrk (alloc $a0 bytes)
	li	$v0, 9
	syscall
	## Restore original memory size
	move	$a0, $t9

	## $a0 holds original memory size, $v0 holds allocated memory offset
	jr	$ra


## Routine: malloc -----------------------------------------
# Call alloc and initialize allocated memory to 0.
# $a0 holds the size of memory wanted.
	.text
malloc:
	## Call alloc (alloc $a0 bytes)
	move	$t1, $ra		# Save return address in $t1
	jal	alloc
	## Fill allocated memory with zeros
	move	$a2, $v0		# $a2 points to allocated memory
	j	Lmalloc_zend
Lmalloc_zbegin:
	sw	$zero, ($a2)
	addi	$a0, $a0, -4
	addi	$a2, $a2, 4		# Next word
Lmalloc_zend:
	bgtz	$a0, Lmalloc_zbegin	# Loop until end of allocated memory

	## $v0 holds allocated and nulled memory offset
	jr	$t1


## Routine: exit -------------------------------------------
# Halt execution
	.text
exit:
	## Call system exit
	li	$v0, 10
	syscall

	
## Routine: init_array -------------------------------------
# $a0 holds the size of the array
# $a1 holds initial value
	.text
init_array:
	## Call sbrk
	sll	$a0, $a0, 2		# Each element is 4 bytes long
	li	$v0, 9
	syscall
	## Fill allocated memory with value in $a1
	move	$a2, $v0		# $a2 points to allocated memory
	j	Linit_array_vend
Linit_array_vbegin:
	sw	$a1, ($a2)
	addi	$a0, $a0, -4
	add	$a2, $a2, 4		# Next word
Linit_array_vend:
	bgtz	$a0, Linit_array_vbegin	# Loop until end of allocated memory

	## $v0 holds allocated and initialized memory offset
	jr	$ra


## Routine: strcmp -----------------------------------------
# Compare string $a0 with string $a1 (normalize result to -1, 0, +1)
	.text
strcmp:
	## Same start address implies equality
	beq	$a0, $a1, Lstrcmp_eq

	## Compute sizes
	lw	$a2, ($a0)		# $a2 holds the size of $a0
	lw	$a3, ($a1)		# $a3 holds the size of $a1
	## Compute start address of characters 
	addiu	$a0, $a0, 4		# $a0 points on fst string's content
	addiu	$a1, $a1, 4		# $a1 points on snd string's content

	## Scan the strings for comparison
Lstrcmp_sbegin:
	## If $a2 or $a3 are not equal to zero, branch on scan body
	bne	$zero, $a2, Lstrcmp_sbody
	bne	$zero, $a3, Lstrcmp_sbody
	
	## Strings are both empty so equal
	j	Lstrcmp_eq
Lstrcmp_sbody:
	
	## If end of fst string $a0 is reached, $a0 is lower than $a1
	beq	$zero, $a2, Lstrcmp_lt
	## If end of snd string $a1 is reached, $a0 is greater than $a1
	beq	$zero, $a3, Lstrcmp_gt
	## Compare two characters
	lbu	$t0, ($a0)
	lbu	$t1, ($a1)
	blt	$t0, $t1, Lstrcmp_lt
	bgt	$t0, $t1, Lstrcmp_gt
	## Characters are equal, continue to scan
	addiu	$a0, $a0, 1
	addiu	$a2, $a2, -1
	addiu	$a1, $a1, 1
	addiu	$a3, $a3, -1
	j	Lstrcmp_sbegin

Lstrcmp_lt:
	## Fst string $a0 is lower than snd $a1, return -1
	li	$v0, -1
	jr	$ra
Lstrcmp_eq:
	## Strings are equal, store 0 in result register
	li	$v0, 0
	jr	$ra
Lstrcmp_gt:
	## Fst string $a0 is greater than snd $a1, return 1
	li	$v0, 1
	jr	$ra
	
## Routine: print ------------------------------------------
# Print the string $a0
	.text
print:
	## Compute start address of characters
	lw	$a1, ($a0)		# $a1 holds the size of $a0
	add	$a0, $a0, 4
	## Null terminate string
	add	$a2, $a0, $a1
	lb	$a3, ($a2)
	sb	$zero, ($a2)		# Store a zero in beyond byte
	## Call system's print string
	li	$v0, 4
	syscall
	## Restore string's end
	sb	$a3, ($a2)
	## Content of $v0 is undetermined
	jr	$ra


## Routine: print_int --------------------------------------
# Print the integer $a0
	.text
print_int:
	## Call system's print int
	li	$v0, 1
	syscall
	## Content of $v0 is undetermined	
	jr	$ra
	
## Routine flush -------------------------------------------
# Nothing to do
	.text
flush:
	jr	$ra


## Routine: main -------------------------------------------
	.data
	.word  0			# Align
Runtconsts:
	.space 2048
Lruntempty:
	.word 0
	
	.text

# 0x004000dc
# Routine main
main:
	li	$a0, 0
	la	$a1, Runtconsts
	li	$a2, 1
Lrunt20:
	sw	$a2, ($a1)
	sb	$a0, 4($a1)
	addiu	$a1, $a1, 8
	addiu	$a0, $a0, 1
	slti	$a3, $a0, 256
	bne	$zero, $a3, Lrunt20
	move	$fp, $sp
	li	$a0, 0
	jal	t_main
	move	$s0, $v0   # save the return value
	li	$a0, 0	   # exit sucessfully
	li	$v0, 10
	syscall

## Routine: ord --------------------------------------------
# Compute ascii value of the first character of string $a0
	.text
ord:
	## Return -1 on empty string
	lw	$a1, ($a0)		# $a1 holds the size of $a0
	## If string $a0 is empty, return -1
	li	$v0, -1
	beq	$zero, $a1, Lord_end
	## Else return first char of $a0 into $v0
	lbu	$v0, 4($a0)
Lord_end:
	jr	$ra


## Routine: chr --------------------------------------------
# Return a one-character string which ascii value is $a0
	.data
Lchr_except_msg:
	.asciiz "chr: character out of range\n"

	.text
chr:
	## Check range of $a0
	blt	$a0, $zero, Lchr_except	# Raise exception on ascii code < 0
	li	$a1, 256
	bge	$a0, $a1, Lchr_except	# Raise exception on ascii code >= 256
	## Return the constant string
	la	$v0, Runtconsts
	sll	$a0, $a0, 3
	add	$v0, $v0, $a0

	## $v0 holds the address of a constant string
	jr	$ra

Lchr_except:
	## Runtime failure.
	la	$a0, Lchr_except_msg
	li	$v0, 4
	syscall
	li	$a0, 120
	li	$v0, 10
	syscall


## Routine: size -------------------------------------------
# Return size of string $a0
	.text
size:
	lw	$v0, ($a0)		# $v0 holds the size of $a0
	jr	$ra


## Routine: substring --------------------------------------
# Return string $a0 extracted from $a1 for $a2 characters
	.data
Lsubstring_except_msg:
	.asciiz "substring: arguments out of bounds\n"
	.text
substring:
	## Preconditions
	bltz	$a1, Lsubstring_except	# Exception on start index < 0
	bltz	$a2, Lsubstring_except	# Exception on start length < 0
	lw	$t1, ($a0)		# $t1 holds the size of $a0
	add	$t2, $a1, $a2		# $t2 holds last character offset 
	bgt	$t2, $t1, Lsubstring_except	# Exception on offset > size

	## Special case for length = 1: use CONSTS
	li	$t3, 1
	bne	$a2, $t3, Lsubstring_non_1
	## One character wanted, use constant strings
	add	$t1, $a0, $a1
	addiu	$t1, 4
	lbu	$a0, ($t1)
	## Hard jump to chr
	j	chr

Lsubstring_non_1:
	## Special case for length = 0: use EMPTY
	bne	$zero, $a2, Lsubstring_allocate
	la	$v0, Lruntempty
	jr	$ra

Lsubstring_allocate:
	## $t1 points to start of wanted characters of $a0
	addiu	$t1, $a0, 4
	add	$t1, $t1, $a1
	## Keep room for the ending `0' required by print.
	addiu	$a0, $a2, 5		## FIXME: useless ?
	## Call alloc
	move	$t4, $ra		# Save caller (return) address
	jal	alloc

	## Fill new string structure
	sw	$a2, ($v0)
	addiu	$t2, $v0, 4		# $t2 holds start of new chars
Lsubstring_cpy:
	## Copy string $a0 from chars $a1 to $a2 into new string
	lbu	$t3, ($t1)
	sb	$t3, ($t2)
	addiu	$t1, 1
	addiu	$t2, 1
	addiu	$a2, -1
	bgtz	$a2, Lsubstring_cpy	# Do loop while size is greater than 0
	## FIXME:	 null-terminate it ?
	sb	$zero, ($t2)

	## Result string is $v0
	jr	$t4

Lsubstring_except:
	li	$v0, 4
	la	$a0, Lsubstring_except_msg
	syscall
	li	$a0, 120
	li	$v0, 10
	syscall
	
## Routine: concat -----------------------------------------
# Concatenate string $a0 with string $a1
concat:
# <<
	## Compute sizes
	lw	$t0, ($a0)		# $t0 holds the size of $a0
	lw	$t1, ($a1)		# $t1 holds the size of $a1
	## Check simple cases (empty string)
	beq	$zero, $t0, Lconcat_ret_snd
	beq	$zero, $t1, Lconcat_ret_fst
	## Compute start address of characters 
	addiu	$t2, $a0, 4		# $t2 holds start of $a0 chars
	addiu	$t3, $a1, 4		# $t3 holds start of $a1 chars
	
	## Compute resulting string size
	add	$t4, $t0, $t1
	## Keep a room for the size and the additional `0' required by print(?)
	addiu	$a0, $t4, 5
	## Call alloc
	move	$t6, $ra		# Save caller (return) address
	jal	alloc

	## Fill new string structure
	addiu	$t5, $v0, 4		# $t5 holds start of new chars
	sw	$t4, ($v0)		# $t4 holds size of new string
Lconcat_fst:
	## Copy first string $a0 into new string
	lbu	$a0, ($t2)
	sb	$a0, ($t5)
	addiu	$t2, 1
	addiu	$t5, 1
	addiu	$t0, -1
	bgtz	$t0, Lconcat_fst	# Do loop while size is greater than 0
Lconcat_snd:
	## Copy second string $a1 into new string
	lbu	$a0, ($t3)
	sb	$a0, ($t5)
	addiu	$t3, 1
	addiu	$t5, 1
	addiu	$t1, -1
	bgtz	$t1, Lconcat_snd	# Do loop while size is greater than 0

	## Result string is $v0
	jr	$t6
	
Lconcat_ret_fst:
	## Second string is empty, return first
	move	$v0, $a0
	jr	$ra
Lconcat_ret_snd:
	## First string is empty, return second
	move	$v0, $a1
	jr	$ra

	
## Routine: _not -------------------------------------------
# Logical negation
	.text
_not:
	beq	$a0, $zero, L_not_1
	li	$v0, 0
	jr	$ra
L_not_1:
	li	$v0, 1
	jr	$ra


## Routine: getchar ----------------------------------------
	.data
getchbuf:
	.space 200
	.word 0			# sentry
	.word 0			# FIXME: useless

getchptr:
	.word getchbuf

	.text
getchar:
	## Load current position in buffer (getchptr) into $a0
	la	$a0, getchptr
	lw	$a0, ($a0)
	## If $a0 does not point to end of buffer, go to Lgetchar_ret
	lb	$v0, ($a0)
	bne	$zero, $v0, Lgetchar_ret

	## End of buffer: try to read 200 chars into buffer (getchbuf)
	la	$a0, getchbuf		# Load destination address
	li	$a1, 200		# Read 200 characters
	li	$v0, 8
	syscall				# Call system

	## Load current position in buffer (getchptr) into $a0
	la	$a0, getchbuf
	
	## If string is not empty, go to Lgetchar_ret
	lbu	$v0, ($a0)
	bne	$zero, $v0, Lgetchar_ret

	## String is still empty, return constant empty string
	la	$v0, Lruntempty
	jr	$ra

Lgetchar_ret:
	## $v0 holds the read character
	## Walk through the buffer
	add	$a0, $a0, 1
	la	$a1, getchptr
	sw	$a0, ($a1)

	## Hard jump to chr to return a constant string equivalent to $v0
	move	$a0, $v0
	j	chr

## End of runtime -------------------------------------------------------------

# == Final assembler ouput. == #
.data
l0: 
	.word 1
	.asciiz "\n"
.text

# Routine: Main
t_main:
	sw	$fp, ($sp)
	move	$fp, $sp
	sub	$sp, $sp, 8
	sw	$ra, -4 ($fp)
l1: 
	li	$t0, 2
	mul	$t1, $t0, 3
	li	$t0, 1
	add	$a0, $t0, $t1
	jal	print_int
	la	$a0, l0
	jal	print
l2: 
	
	lw	$ra, -4 ($fp)
	move	$sp, $fp
	lw	$fp, ($fp)
	jr	$ra
